//
//  SampleRender.m
//  opengl_texture
//
//  Created by German Saprykin on 22/4/18.
//

#import "NormalImageRenderWorker.h"
#import <SDWebImage/SDWebImage.h>
#import <SDWebImageWebPCoder.h>
#import "GLKit/GLKit.h"

/**
 定义顶点类型
 */
typedef struct {
    GLKVector3 positionCoord; // (X, Y, Z)
    GLKVector2 textureCoord; // (U, V)
} SenceVertex;

@interface NormalImageRenderWorker () {
    GLuint _program;
}

@property (nonatomic, copy) NSString *url;
@property (nonatomic, assign) CGFloat aspectRatio;
@property (nonatomic, assign) CGFloat width;
@property (nonatomic, assign) CGFloat height;
@property (nonatomic, assign) BOOL centerCrop;

@property (nonatomic, assign) SenceVertex *vertices; // 顶点数组
@property (nonatomic, strong) UIImage *image;
@property (nonatomic, strong) NSArray *imageArray;
@property (nonatomic, assign) CGFloat duration;
@property (nonatomic, strong) NSDate *startTime;

@end

@implementation NormalImageRenderWorker

- (instancetype)initWithCall:(FlutterMethodCall *)call {
    self = [super init];

    if (self) {
        [self initCallArgument:call];
    }

    return self;
}

- (void)initCallArgument:(FlutterMethodCall *)call {
    SDImageWebPCoder *webPCoder = [SDImageWebPCoder sharedCoder];
    [[SDImageCodersManager sharedManager] addCoder:webPCoder];

    _url = call.arguments[@"url"];
    _aspectRatio = [call.arguments[@"aspectRatio"] floatValue];
    _width = [call.arguments[@"width"] floatValue];
    _height = [call.arguments[@"height"] floatValue];
    _centerCrop = [call.arguments[@"centerCrop"] boolValue];

    NSLog(@"call Argument: \n url: %@, \n aspectRatio: %.1f, \n width:%.1f, \n height:%.1f, \n centerCrop:%ld",
            self.url, self.aspectRatio, self.width, self.height, (long) self.centerCrop);
}

- (void)onCreate {
    self.image = nil;
    self.imageArray = nil;
    self.duration = 0;

    UIImage *image;

    image = [[SDImageCache sharedImageCache] imageFromCacheForKey:self.url];
    if (!image) {
        [[SDWebImageDownloader sharedDownloader] downloadImageWithURL:[[NSURL alloc] initWithString:self.url] completed:^(UIImage *image, NSData *data, NSError *error, BOOL finished) {
            NSLog(@"Download image done: %@", self.url);
            dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
                if (!error && image) {
                    self.image = image;
                    self.imageArray = image.images;
                    self.duration = image.duration;
                    self.startTime = [NSDate date];

                    // GIF图片保存后再使用会丢失images数据，无法动起来
                    if (image.sd_imageFormat != SDImageFormatGIF) {
                        [[SDImageCache sharedImageCache] storeImage:image forKey:self.url completion:nil];
                    }
                }
            });
        }];
        
    } else {
        NSLog(@"cache image found: %@",self.url);
        self.image = image;
        self.imageArray = image.images;
        self.duration = image.duration;
        self.startTime = [NSDate date];
    }

    // 创建顶点数组
    self.vertices = malloc(sizeof(SenceVertex) * 4); // 4 个顶点

    self.vertices[0] = (SenceVertex) {{-1, 1, 0}, {0, 0}}; // 左上角
    self.vertices[1] = (SenceVertex) {{-1, -1, 0}, {0, 1}}; // 左下角
    self.vertices[2] = (SenceVertex) {{1, 1, 0}, {1, 0}}; // 右上角
    self.vertices[3] = (SenceVertex) {{1, -1, 0}, {1, 1}}; // 右下角

    // 编译链接 shader
    _program = [self programWithShaderName:@"glsl"]; // glsl.vsh & glsl.fsh

}

// 将一个顶点着色器和一个片段着色器挂载到一个着色器程序上，并返回程序的 id
- (GLuint)programWithShaderName:(NSString *)shaderName {
    // 编译两个着色器
    GLuint vertexShader = [self compileShaderWithName:shaderName type:GL_VERTEX_SHADER];
    GLuint fragmentShader = [self compileShaderWithName:shaderName type:GL_FRAGMENT_SHADER];

    // 挂载 shader 到 program 上
    GLuint program = glCreateProgram();
    glAttachShader(program, vertexShader);
    glAttachShader(program, fragmentShader);

    // 链接 program
    glLinkProgram(program);

    // 检查链接是否成功
    GLint linkSuccess;
    glGetProgramiv(program, GL_LINK_STATUS, &linkSuccess);
    if (linkSuccess == GL_FALSE) {
        GLchar messages[256];
        glGetProgramInfoLog(program, sizeof(messages), 0, &messages[0]);
        NSString *messageString = [NSString stringWithUTF8String:messages];
        NSAssert(NO, @"program链接失败：%@", messageString);
        exit(1);
    }
    return program;
}

// 编译一个 shader，并返回 shader 的 id
- (GLuint)compileShaderWithName:(NSString *)name type:(GLenum)shaderType {
    // 查找 shader 文件
    NSString *bundlePath = [[NSBundle bundleForClass:[self class]].resourcePath
            stringByAppendingPathComponent:@"/TipFlutterImageView.bundle"];
    NSBundle *bundle = [NSBundle bundleWithPath:bundlePath];
    NSString *shaderPath = [bundle pathForResource:name ofType:shaderType == GL_VERTEX_SHADER ? @"vsh" : @"fsh"]; // 根据不同的类型确定后缀名
    NSError *error;
    NSString *shaderString = [NSString stringWithContentsOfFile:shaderPath encoding:NSUTF8StringEncoding error:&error];
    if (!shaderString) {
        NSAssert(NO, @"读取shader失败");
        exit(1);
    }

    // 创建一个 shader 对象
    GLuint shader = glCreateShader(shaderType);

    // 获取 shader 的内容
    const char *shaderStringUTF8 = [shaderString UTF8String];
    int shaderStringLength = (int) [shaderString length];
    glShaderSource(shader, 1, &shaderStringUTF8, &shaderStringLength);

    // 编译shader
    glCompileShader(shader);

    // 查询 shader 是否编译成功
    GLint compileSuccess;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &compileSuccess);
    if (compileSuccess == GL_FALSE) {
        GLchar messages[256];
        glGetShaderInfoLog(shader, sizeof(messages), 0, &messages[0]);
        NSString *messageString = [NSString stringWithUTF8String:messages];
        NSAssert(NO, @"shader编译失败：%@", messageString);
        exit(1);
    }

    return shader;
}

- (BOOL)onDraw {
    if (!self.image) {
        return NO;
    }

    [self draw];

    return YES;
}

- (void)draw {
    // 设置视口尺寸
    glViewport(0, 0, (GLsizei) self.width, (GLsizei) self.height);

    // 读取纹理
    GLuint textureId = [self createTextureWithImage:[self getCurrentImage]];

    glUseProgram(_program);

    // 获取 shader 中的参数，然后传数据进去
    GLuint positionSlot = (GLuint) glGetAttribLocation(_program, "Position");
    GLuint textureSlot = (GLuint) glGetUniformLocation(_program, "Texture");  // 注意 Uniform 类型的获取方式
    GLuint textureCoordsSlot = (GLuint) glGetAttribLocation(_program, "TextureCoords");

    // 将纹理 ID 传给着色器程序
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, textureId);
    glUniform1i(textureSlot, 0);  // 将 textureSlot 赋值为 0，而 0 与 GL_TEXTURE0 对应，这里如果写 1，上面也要改成 GL_TEXTURE1

    // 创建顶点缓存
    GLuint vertexBuffer;
    glGenBuffers(1, &vertexBuffer);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
    GLsizeiptr bufferSizeBytes = sizeof(SenceVertex) * 4;
    glBufferData(GL_ARRAY_BUFFER, bufferSizeBytes, self.vertices, GL_STATIC_DRAW);

    // 设置顶点数据
    glEnableVertexAttribArray(positionSlot);
    glVertexAttribPointer(positionSlot, 3, GL_FLOAT, GL_FALSE, sizeof(SenceVertex), NULL + offsetof(SenceVertex, positionCoord));

    // 设置纹理数据
    glEnableVertexAttribArray(textureCoordsSlot);
    glVertexAttribPointer(textureCoordsSlot, 2, GL_FLOAT, GL_FALSE, sizeof(SenceVertex), NULL + offsetof(SenceVertex, textureCoord));

    // 开始绘制
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    // 删除顶点缓存
    glDeleteBuffers(1, &vertexBuffer);
    vertexBuffer = 0;
    glBindTexture(GL_TEXTURE_2D, 0);
}

// 通过一张图片来创建纹理
- (GLuint)createTextureWithImage:(UIImage *)image {
    // 将 UIImage 转换为 CGImageRef
    CGImageRef cgImageRef = [image CGImage];
    GLuint width = (GLuint) CGImageGetWidth(cgImageRef);
    GLuint height = (GLuint) CGImageGetHeight(cgImageRef);
    CGRect rect = CGRectMake(0, 0, width, height);

    // 绘制图片
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    void *imageData = malloc(width * height * 4);
    CGContextRef context = CGBitmapContextCreate(imageData, width, height, 8, width * 4, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big);
    CGContextTranslateCTM(context, 0, height);
    CGContextScaleCTM(context, 1.0f, -1.0f);
    CGColorSpaceRelease(colorSpace);
    CGContextClearRect(context, rect);
    CGContextDrawImage(context, rect, cgImageRef);

    // 生成纹理，不能每次都glGenTextures，会导致内存泄漏，这里只能用0（default），试过1（需要通过glGenTextures）会黑屏
//    GLuint textureID;
//    glGenTextures(1, &textureID);
    glBindTexture(GL_TEXTURE_2D, 0);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, imageData); // 将图片数据写入纹理缓存

    // 设置如何把纹素映射成像素
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    // 解绑
    glBindTexture(GL_TEXTURE_2D, 0);

    // 释放内存
    CGContextRelease(context);
    free(imageData);

    return 0;
}

- (void)onDispose {
    if (_vertices) {
        free(_vertices);
        _vertices = nil;
    }
}

- (void)setImage:(UIImage *)image {
    if (image && self.centerCrop) {
        image = [self imageByCroppingImage:image toSize:CGSizeMake(self.width, self.height)];
    }

    _image = image;
}

- (UIImage *)getCurrentImage {
    UIImage *image = self.image;
    if (self.imageArray && self.imageArray.count > 0 && self.duration > 0.01) {
        NSTimeInterval delta = [[NSDate date] timeIntervalSinceDate:self.startTime];
        NSUInteger index = (NSUInteger) (((int) (delta * 1000) % (int) (self.duration * 1000)) * 1.0f / (int) (self.duration * 1000) * self.imageArray.count);
        image = self.imageArray[index];
    }

    return image;
}

- (UIImage *)imageByCroppingImage:(UIImage *)image toSize:(CGSize)size {
    double newCropWidth, newCropHeight;

    //=== To crop more efficently =====//
    if (image.size.width < image.size.height) {
        if (image.size.width < size.width) {
            newCropWidth = size.width;
        } else {
            newCropWidth = image.size.width;
        }
        newCropHeight = (newCropWidth * size.height) / size.width;
    } else {
        if (image.size.height < size.height) {
            newCropHeight = size.height;
        } else {
            newCropHeight = image.size.height;
        }
        newCropWidth = (newCropHeight * size.width) / size.height;
    }

    double x = image.size.width / 2.0 - newCropWidth / 2.0;
    double y = image.size.height / 2.0 - newCropHeight / 2.0;

    CGRect cropRect = CGRectMake(x, y, newCropWidth, newCropHeight);
    CGImageRef imageRef = CGImageCreateWithImageInRect([image CGImage], cropRect);

    UIImage *cropped = [UIImage imageWithCGImage:imageRef];
    CGImageRelease(imageRef);

    return cropped;
}

@end
